/*
 * CSetFinder.java
 *
 * This file contains the main functions for the CRANS
 *  cross rotation analysis algorithm as described in:
 *
 *  "A Subgroup Algorithm to Identify Cross-Rotation Peaks
 *   Consistent with Non-Crystallographic Symmetry." R. H. Lilien,
 *   C. Bailey-Kellogg, A. C. Anderson, and B. R. Donald, Acta
 *   Crystallographica D, 60:1057-67 (2004).
 *
 * Ryan Lilien
 * Chris Bailey-Kellogg

////////////////////////////////////////////////////////////////////////////////////////////
//  CSetFinder.java
//
//  Version:           0.1
//
//
// authors:
//    initials    name            organization                email
//   ---------   --------------  ------------------------    ------------------------------
//     RHL        Ryan Lilien     Dartmouth College           lilien@cs.dartmouth.edu
//
////////////////////////////////////////////////////////////////////////////////////////////

*/

package CRANS;

import java.util.*;
import java.io.*;

public class CSetFinder {
    
    // Default empty constructor
    public CSetFinder() {
    }

    // Display usage information
    public void dispUsage() {
        System.out.println("");
        System.out.println("-----------------------------------------------\\");
        System.out.println("|                                              |");
        System.out.println("| Cross-Rotation ANalysis of Symmetry (CRANS)  |");
        System.out.println("|   Version 1.2                                |");
        System.out.println("-----------------------------------------------/");
        System.out.println();
        System.out.println(" PAPER REF ");
        System.out.println();
        System.out.println(" CRANS Usage...");
        System.out.println();
        System.out.println("  Input parameters:");
        System.out.println("    If not specified by the user, input parameters assume default values");
        System.out.println();
        System.out.println("   -i string\n\tstring: Cross-rotation input filename (ie. crossrot.list)");
        System.out.println("   -o string\n\tstring: Output filename (default cross_rot_NCS.txt)");
        System.out.println("   -x float\n\tfloat: Axis tolerance (degrees) (default 4.5)");
        System.out.println("   -r float\n\tfloat: Rotation tolerance (degrees) (default 5)");
        System.out.println("   -z integer\n\tinteger: Max number of results to print (default 20)");
        System.out.println("   -d string char char\n\tUsed when the molecular replacement model is a dimer");
        System.out.println("\tstring: PDB filename (ie. molmodel.pdb)");
        System.out.println("\tchar: chain 1 id (ie. A)");
        System.out.println("\tchar: chain 2 id (ie. B)");
        System.out.println("   -a float float float\n\tfloats: Self-rotation NCS axis (in Euclidian coords (x,y,z))");
        System.out.println("\tNOTE: If an axis is present then results must be within axis tolerance of it");
        System.out.println("   -p\n\tPrint (display) all results (if not specified, subsets are not displayed)");
        System.out.println("   -n integer\n\tUsed in 'normal' mode (not scan mode)");
        System.out.println("\tNOTE: You can not use both -n and -s at the same time");
        System.out.println("\tinteger: Degree of NCS (default 5)");
        System.out.println("   -s integer integer\n\tUsed in 'scan' mode, all sets of rotations from low to high are returned");
        System.out.println("\tNOTE: You can not use both -n and -s at the same time");
        System.out.println("\tinteger: low NCS (ie. 3)");
        System.out.println("\tinteger: high NCSB (ie. 8)");
        System.out.println("\tNOTE: You can not use both -n and -s at the same time");
        System.out.println("   -m integer\n\tinteger: Max allowable number of missing rotation peaks per set (default 0)");
        System.out.println("   -c \n\tUse axis-angle deviation scoring rather than cross-rotation scoring");
        System.out.println();
        System.out.println();
        System.exit(0);
    }

    // Displays a specific error message and then exits
    public void dispErrorParams(String cmd) {
        System.out.println();
        System.out.println("Error: Invalid parameters for option "+cmd);
        System.exit(0);
    }

    // Where all the magic happens
    public void findCSets(String[] args) {

        // Default Arguments
        double rot_tol = 5.0;
        double axis_tol = 4.5;
        int maxMissing = 0;
        int min_group_sz = 0;
        int print_length = 20;
        boolean isDimer = false;
        String dname = null;  // Name of the dimer file (pdb)
        String chainA = null; // Chain A id (for dimer case)
        String chainB = null; // Chain B id (for dimer case)
        boolean printAll = false;
        int lowSym = 5;
        int highSym = 5;
        int nSym = 0;
        String fname = null;  // Name of the cross-rot results file
        String outname = "cross_rot_NCS.txt";  // Output filename
        double[] axs = null;
        PrintStream outfile = null;
        boolean RFScoring = true;
        boolean selfAxisPresent = false;
        double NCSx = 0.0, NCSy = 0.0, NCSz = 0.0;
        
        // Parse input parameters
        int a = 0;
        while (args.length > a) {
            if (args[a].equalsIgnoreCase("-p")) {
                printAll = true;
                a++;
            } else if (args[a].equalsIgnoreCase("-i")) {
                a++;
                if (a < args.length)
                    fname = args[a++];
                else
                    dispErrorParams("-i");
            } else if (args[a].equalsIgnoreCase("-o")) {
                a++;
                if (a < args.length)
                    outname = args[a++];
                else
                    dispErrorParams("-o");
            } else if (args[a].equalsIgnoreCase("-x")) {
                a++;
                if (a < args.length)
                    axis_tol = (new Double(args[a++]).doubleValue());
                else
                    dispErrorParams("-x");
            } else if (args[a].equalsIgnoreCase("-r")) {
                a++;
                if (a < args.length)
                    rot_tol = (new Double(args[a++]).doubleValue());
                else
                    dispErrorParams("-r");
            } else if (args[a].equalsIgnoreCase("-z")) {
                a++;
                if (a < args.length)
                    print_length = (new Integer(args[a++]).intValue());
                else
                    dispErrorParams("-z");
            } else if (args[a].equalsIgnoreCase("-c")) {
                a++;
                RFScoring = false;
            } else if (args[a].equalsIgnoreCase("-d")) {
                a++;
                if (a+2 < args.length) {
                    isDimer = true;
                    dname = args[a++];
                    chainA = args[a++];
                    chainB = args[a++];
                } else
                    dispErrorParams("-d");
            } else if (args[a].equalsIgnoreCase("-n")) {
                a++;
                if (a < args.length) {
                    lowSym = (new Integer(args[a++]).intValue());
                    highSym = lowSym;
                } else
                    dispErrorParams("-n");
            } else if (args[a].equalsIgnoreCase("-s")) {
                a++;
                if (a+1 < args.length) {
                    lowSym = (new Integer(args[a++]).intValue());
                    highSym = (new Integer(args[a++]).intValue());
                } else
                    dispErrorParams("-s");
            } else if (args[a].equalsIgnoreCase("-a")) {
                a++;
                if (a+2 < args.length) {
                    NCSx = (new Double(args[a++]).doubleValue());
                    NCSy = (new Double(args[a++]).doubleValue());
                    NCSz = (new Double(args[a++]).doubleValue());
                    double tmpd = Math.sqrt(NCSx*NCSx + NCSy*NCSy + NCSz*NCSz);
                    if (tmpd == 0.0) {
                        System.out.println("ERROR: Self-Rotation NCS axis can not be zero");
                        System.exit(0);
                    }
                    NCSx = NCSx / tmpd;
                    NCSy = NCSy / tmpd;
                    NCSz = NCSz / tmpd;
                    selfAxisPresent = true;
                } else
                    dispErrorParams("-a");
            } else if (args[a].equalsIgnoreCase("-m")) {
                a++;
                if (a < args.length)
                    maxMissing = (new Integer(args[a++]).intValue());
                else
                    dispErrorParams("-m");
            }
        }

        if (fname == null)
            dispUsage();

        if(lowSym < 2){
            System.out.println("Error: Minimum symmetry group size is 2");
            System.exit(0);
        }
        if(lowSym > highSym){
            System.out.println("Error: Low NCS must be less than or equal to High NCS");
            System.exit(0);
        }
        if((lowSym == 2) && (isDimer)){
            System.out.println("Caution: A 2-fold NCS search involving a dimer may find the dimer axis in addition to NCS symmetry");
        }
        
        // Function variables
        Vector rots = new Vector();
        Vector masterRots = new Vector();
        int numRots = 0;
        int masterNumRots = 0;
        Quaternion dimerQuat = null;
        int[] numSetsFound = new int[highSym+1];

        // Open output file
        try{
            FileOutputStream fileOutputStream = new FileOutputStream(outname);
            BufferedOutputStream bufferedOutputStream = new BufferedOutputStream(fileOutputStream);
            outfile = new PrintStream(bufferedOutputStream);
        }
        catch(Exception e){
            System.out.println("ERROR opening output file");
            System.exit(0);
        }

        
        if (isDimer) {
            System.out.println("Reading dimer file: "+dname);
            GetDimerAxis gda = new GetDimerAxis(dname,chainA,chainB);
            axs = gda.computeAxis();
            if (axs == null) {
                System.out.println("Unable to compute dimer axis from file "+dname);
                System.exit(0);
            }
            System.out.println("Dimer axis found: "+axs[0]+" "+axs[1]+" "+axs[2]);
            dimerQuat = new Quaternion();
            dimerQuat.fromAxisAngle(axs,180.0);
        }

        
        outfile.println("Cross Rotation NCS Analysis of: "+fname);
        outfile.println("");
        outfile.println("Search Parameters:");
        outfile.println(" Input filename: "+fname);
        outfile.println(" Axis tol: "+axis_tol);
        outfile.println(" Rotation tol: "+rot_tol);
        outfile.println(" Max results to display: "+print_length);
        if (isDimer) {
            outfile.println(" Molecular replacement model is a dimer: "+dname);
            outfile.println("  Chains: "+chainA+" "+chainB);
            outfile.println("  Dimer axis: "+axs[0]+" "+axs[1]+" "+axs[2]);
        }
        if (lowSym == highSym) {
            outfile.println(" Symmetry search mode: Normal mode");
            outfile.println(" NCS-fold symmetry: "+lowSym);
        } else {
            outfile.println(" Symmetry search mode: Scan mode");
            outfile.println(" NCS-fold symmetry range: "+lowSym+" to "+highSym);
        }
        outfile.println(" Maximum number of missing rotation peaks: "+maxMissing);
        if (RFScoring)
            outfile.println(" Rotation Function scores used to score groups (higher = better)");
        else
            outfile.println(" Axis-Angle deviation used to score groups (lower = better)");
        outfile.println("");
        outfile.println("");

        // Read cross-rotation list
        masterNumRots = readCrossRotationList(fname,masterRots);
        System.out.println("Number of cross-rotation peaks read: "+masterNumRots);

        // If we're dealing with a dimer, generate the
        //  extra rotations
        if (isDimer){
            masterRots = generateDimerRelatedRots(masterRots, dimerQuat);
            masterNumRots = masterRots.size();
        }

        
        // Begin the scan (if only one NCS-degree is being tested
        //  then lowSym==highSym)
        for(int curSym=lowSym;curSym<=highSym;curSym++) {

            // Grab a copy of the main rotation vector
            rots = (Vector)masterRots.clone();
            numRots = masterNumRots;
            min_group_sz = curSym - maxMissing;
            if (min_group_sz < 2)
                min_group_sz = 2;
            
            System.out.println("Starting search for "+curSym+"-fold symmetry");
            
            // Compute all pairwise differences
            RotDifference[][] allDiffs = new RotDifference[numRots][numRots];
            System.out.print(" Computing all pairwise rotation differences ...");
            computeAllPairsRotationDiffs(rots,allDiffs,curSym);
            System.out.println("done");
        
            // Remove differences whose angles are not within
            //  tol of a multiple of 360/curSym
            System.out.print(" Finding NCS related rotation sets ...");
            findCloseDifferences(allDiffs,numRots,rot_tol);
                        
            // Group similar differences, at the same time, if we have a
            //  self-rotation NCS axis, remove differences that are
            //  not within axis_tol of the NCS axis
            Vector rotationSets = new Vector();
            rotationSets = computeRotationSets(rotationSets,allDiffs,numRots,curSym,
                                    rot_tol,axis_tol,min_group_sz,RFScoring,
                                    selfAxisPresent,NCSx,NCSy,NCSz);
            System.out.println("done");

            // Fill in holes
            System.out.print(" Patching sets with missing rotation peaks ...");
            fillHoles(rotationSets,allDiffs,curSym);
            System.out.println("done");

            // Remove subsets
            if (!printAll){
                System.out.print(" Removing subsets ...");
                rotationSets = removeSubsets(rotationSets);
                System.out.println("done");
            }

            // Sort sets
            System.out.print(" Sorting and outputting sets ...");
            Object[] finalRSets = rotationSets.toArray();
            Arrays.sort(finalRSets);

            // Print sets
            for(int i=0;(i<finalRSets.length) && (i<print_length);i++)
                printSet(outfile,(RotSet)finalRSets[i],allDiffs,i,curSym);
            System.out.println("done");
            
            numSetsFound[curSym] = Math.min(finalRSets.length,print_length);
        }
        
        outfile.println("");
        outfile.println("");
        for(int q=lowSym;q<=highSym;q++)
            outfile.println(q+"-fold sets found: "+numSetsFound[q]);
        outfile.close();
    }   

    
    // Removes subsets and sets with more flips
    public Vector removeSubsets(Vector rSets) {
        
        Vector newRSets = new Vector();
        RotSet rSetA = null;
        boolean accept = true;
        int w=0;
        
        for(int q=0;q<rSets.size();q++){
            rSetA = (RotSet)rSets.get(q);
            accept = true;
            w=0;
            while (accept && (w<rSets.size())) {
                if (isSubset(rSetA,(RotSet)rSets.get(w)))
                    accept = false;
                w++;
            }
            if (accept)
                newRSets.add(rSetA);
        }
        return(newRSets);
    }
    
    
    // Returns true if rSetA is a subset of rSetB
    // A is a subset of B if every non-simulated rotation in
    //  A is a non-simulated rotation in B AND if A has fewer
    //  dimer-flips than B.
    public boolean isSubset(RotSet rSetA, RotSet rSetB) {
        
        boolean found = false;
        int w=0;
        Rotation r1 = null, r2 = null;
        int rots1 = 0, rots2 = 0;
        int flip1 = 0, flip2 = 0;
        
        for(int q=0;q<rSetA.rots.size();q++){
            r1 = (Rotation)rSetA.rots.get(q);
            if (!r1.simulated) {
                rots1++;
                if (r1.dimerRotation)
                    flip1++;
                w=0;
                found=false;
                while (!found && (w<rSetB.rots.size())){
                    r2 = (Rotation)rSetB.rots.get(w);
                    if (!r2.simulated) {
                        if (r1.origString.equalsIgnoreCase(r2.origString))
                            found = true;
                    }
                    w++;
                }
                if (!found)
                    return(false);
            }
        }
        
        // At this point we know that all non-simulated rotations of A
        //  are non-simulated rotations in set B. Now we count the number
        //  of nonsimulated rotation and number of dimerflips in each set.
        // We already have the number of nonsimulated rotations and
        //  dimerflips in A, now count the number in B.
        for(int q=0;q<rSetB.rots.size();q++){
            r2 = (Rotation)rSetB.rots.get(q);
            if (!r2.simulated) {
                rots2++;
                if (r2.dimerRotation)
                    flip2++;
            }
        }
        
        // If the number of non-simulated rotations in A is less than
        //  that of B then A is a subset of B.
        if (rots1 < rots2)
            return(true);
        // At this point it must be the case that rots1 = rots2 since
        //  all rots in A are in B and rots1 >= rots2. So we examine
        //  the number of flips
        if (flip1 > flip2)
            return(true);
        
        return(false);
    }

    
    // Prints a set to the stream outstream
    public void printSet(PrintStream outstream, RotSet rSet, RotDifference[][] diffs, 
                            int num, int nSym){

        Rotation r1 = null;
        Rotation r2 = null;
        outstream.println("Set number: "+(num+1));
        outstream.println(nSym+"-fold symmetry");
        outstream.println("Number of cross-rotation peaks: "+rSet.numTrueRotations());
        outstream.println("Number of computed cross-rotation peaks: "+(nSym-rSet.numTrueRotations()));
        if (rSet.score < 0)
            outstream.println("Set score: "+(-rSet.score));
        else
            outstream.println("Set score: "+rSet.score);
        outstream.println("Axis Error - Mean: "+rSet.mean_axis_err+" Max: "+rSet.max_axis_err);
        outstream.println("Angle Error - Mean: "+rSet.mean_angle_err+" Max: "+rSet.max_angle_err);
        outstream.println("Rotations: (Euler (z,x,z) Representation)");
        for(int i=0;i<nSym;i++){
            r1 = (Rotation)rSet.get(i);
            if (r1.dimerRotation)
                outstream.println(" "+(i+1)+" ("+r1.rot.toEulerString()+") DIMER FLIP of original \""+r1.origString+"\"");
            else if (r1.simulated)
                outstream.println(" "+(i+1)+" ("+r1.rot.toEulerString()+") SIMULATED");
            else
                outstream.println(" "+(i+1)+" ("+r1.rot.toEulerString()+") from original \""+r1.origString+"\"");
        }
        outstream.println("Rotation Differences ((Axis) Angle) Form");
        for(int i=0;i<nSym;i++){
            r1 = (Rotation) rSet.get(i);
            for(int j=(i+1);j<nSym;j++){
                r2 = (Rotation) rSet.get(j);
                if (r1.simulated || r2.simulated) {
                    outstream.println(" "+(i+1)+"-"+(j+1)+" *");
                } else {
                    double[] axis = diffs[r1.index][r2.index].axis;
                    outstream.println(" "+(i+1)+"-"+(j+1)+" (("+axis[0]+" "+axis[1]+" "+axis[2]+") "+(diffs[r1.index][r2.index].angle)+")");
                }
            }
        }
        outstream.println();

    }
    
    
    // Computes the missing rotations for each set
    public void fillHoles(Vector rotationSets, RotDifference[][] diffs,
                                int nSym){
                                    
        int k=0;
        Rotation r1, r2;
        double sign = 1.0;
        RotDifference diff = null;
        double da = 360.0 / (double) nSym;
        Rotation tmpRot = null;
        for(int i=0;i<rotationSets.size();i++){
            RotSet rSet = (RotSet) rotationSets.get(i);
            // Find a diff axis (they should pretty much all be the same)
            // There is always an r1 in position 0, (the first rotation
            //  found is always placed there)
            r1 = (Rotation) rSet.get(0);
            k=1;
            while(rSet.get(k) == null)
                k++;
            r2 = (Rotation) rSet.get(k);
            diff = diffs[r1.index][r2.index];
            sign = 1.0;
            if(diff.nearestSymNumber != k)
                sign = -1.0;
            // For each missing rotation, create one from original rotation
            //  r1 followed by rotation around diff axis by appropriate
            //  multiple.
            for(int j=1;j<nSym;j++){
                if (rSet.get(j) == null){
                    Quaternion q1 = new Quaternion();
                    double angle = sign*j*da;
                    q1.fromAxisAngle(diff.axis,angle);
                    Quaternion q2 = new Quaternion();
                    q2 = q1.multiply(r1.rot);
                    tmpRot = new Rotation(q2);
                    tmpRot.simulated = true;
                    rSet.set(j,tmpRot);
                }
            }
        }
    }
    

    // If baseDiff and rDiff's have opposite axes then adjust the angle number
    public int num_wrt(RotDifference baseDiff, RotDifference rDiff, int nSym){
        
        if ((baseDiff.axis[0]*rDiff.axis[0] +
             baseDiff.axis[1]*rDiff.axis[1] +
             baseDiff.axis[2]*rDiff.axis[2]) < 0.0)
            return(nSym - rDiff.nearestSymNumber);
        return(rDiff.nearestSymNumber);
    }
    
    // Called by makeGroups, given a small set of rotation differences
    //  that have similar rotation axes, this function makes all possible
    //  sets of rotationDifferences.
    // Returned is a vector of original rotations (not differences)
    public Vector formGroups(RotSet rotSet, RotDifference[][] diffs,
                                int nSym, int min_group_sz) {
        
        RotDifference baseDiff = (RotDifference) rotSet.get(0);
        RotDifference tmpDiff = null;
        Vector theGroups = new Vector();  // The vector of RotSets
        int num = 0;
        int sz = 0;
        
        // Make sure the group will be complete
        int[] filled = new int[nSym];
        for(int i=1;i<nSym;i++) filled[i]=0;
        for(int i=0;i<rotSet.size();i++) {
            num = num_wrt(baseDiff,(RotDifference)rotSet.get(i),nSym);
            filled[num] = 1;
        }
        int numFilled=1;
        for(int i=1;i<nSym;i++) numFilled += filled[i];
        if (numFilled < min_group_sz)
            return(new Vector());

        RotSet curRSet = null;
        RotSet tmpRSet = new RotSet();
        tmpRSet.setSize(nSym);
        tmpRSet.set(0,baseDiff.rot1);

        theGroups.add(tmpRSet);
        
        for(int i=0;i<rotSet.size();i++){
            tmpDiff = (RotDifference) rotSet.get(i);
            num = num_wrt(baseDiff,tmpDiff,nSym);
            sz = theGroups.size();
            for(int j=0;j<sz;j++){
                curRSet = (RotSet) theGroups.get(j);
                if((curRSet.get(num)) == null) {
                    curRSet.set(num,tmpDiff.rot2);
                } else {
                    tmpRSet = new RotSet(curRSet);
                    tmpRSet.set(num,tmpDiff.rot2);
                    theGroups.add(tmpRSet);
                }
            }
        }
        
        return(theGroups);
    }
    
    
    // Computes a score for each rotation set
    public void scoreRotSet(RotSet rSet, RotDifference[][] diffs, boolean RFScoring,
                             double NCSx, double NCSy, double NCSz) {

        int nangles = 0;
        int naxes = 0;
        double axis_err;
        double self_err;
        Rotation r1 = null;
        Rotation r2 = null;
        Rotation r3 = null;
        RotDifference tmpDiff12 = null;
        RotDifference tmpDiff23 = null;
        double[] avgaxis = new double[3];
        double axislen = 0.0;
        
        rSet.score = 0.0;
        rSet.mean_axis_err = 0.0;
        rSet.max_axis_err = 0.0;
        rSet.mean_angle_err = 0.0;
        rSet.max_angle_err = 0.0;

        avgaxis[0] = 0.0;
        avgaxis[1] = 0.0;
        avgaxis[2] = 0.0;
        
        // First compute average axis
        for(int i=0;i<rSet.size();i++){
            r1 = (Rotation) rSet.get(i);
            if (r1 != null) {
                for(int j=(i+1);j<rSet.size();j++){
                    r2 = (Rotation) rSet.get(j);
                    if (r2 != null) {
                        tmpDiff12 = diffs[r1.index][r2.index];
                        nangles++;
                        for(int k=0;k<3;k++)
                            avgaxis[k] += tmpDiff12.axis[k];
                    }
                }
            }
        }

        // Normalize the average axis
        axislen = Math.sqrt(avgaxis[0]*avgaxis[0] + 
                            avgaxis[1]*avgaxis[1] + 
                            avgaxis[2]*avgaxis[2]);
        for(int k=0;k<3;k++)
            avgaxis[k] = avgaxis[k] / axislen;
        
        // Then compute deviation angle for each pair to the average
        for(int i=0;i<rSet.size();i++){
            r1 = (Rotation) rSet.get(i);
            if (r1 != null) {
                for(int j=(i+1);j<rSet.size();j++){
                    r2 = (Rotation) rSet.get(j);
                    if (r2 != null) {
                        tmpDiff12 = diffs[r1.index][r2.index];
                        rSet.mean_angle_err += tmpDiff12.error;
                        nangles++;
                        rSet.max_angle_err =
                                Math.max(rSet.max_angle_err, tmpDiff12.error);
                        self_err = 180.0/3.1415926 * Math.acos(
                                        Math.abs(tmpDiff12.axis[0] * NCSx +
                                                 tmpDiff12.axis[1] * NCSy +
                                                 tmpDiff12.axis[2] * NCSz));
                        if (self_err > rSet.max_self_axis_err)
                            rSet.max_self_axis_err = self_err;

                        axis_err = 180.0/3.1415926 * Math.acos(
                                    Math.abs( tmpDiff12.axis[0] * avgaxis[0] +
                                              tmpDiff12.axis[1] * avgaxis[1] +
                                              tmpDiff12.axis[2] * avgaxis[2]));
                        rSet.mean_axis_err += axis_err;
                        naxes++;
                        rSet.max_axis_err = Math.max(
                              rSet.max_axis_err, axis_err);
                    }
                }
            }
        }
        
        if (nangles > 0)
            rSet.mean_angle_err /= nangles;
        if (naxes > 0)
            rSet.mean_axis_err /= naxes;

        if(RFScoring){
            float setScore = 0.0f;
            for(int i=0;i<rSet.size();i++){
                r1 = (Rotation) rSet.get(i);
                if (r1!=null)
                    setScore += r1.quality;
            }
            rSet.score = setScore;
        } else {
            rSet.score = -(rSet.mean_angle_err + rSet.mean_axis_err);
        }
    }
    
    // Create groups of rotations with similar axes
    public Vector makeGroups(Vector rotationDiffSets,RotDifference[][] diffs,
                            int row, int numRots, int nSym, double rot_tol,
                            double axis_tol, int min_group_sz, boolean RFScoring,
                            boolean selfAxisPresent, double NCSx, double NCSy,
                            double NCSz) {
        
        Vector tmpGroups = null;  // groups of rotations
        Vector goodGroups = new Vector();
        RotSet tmpRotSet = null;
        
        for(int i=0;i<rotationDiffSets.size();i++){
            tmpGroups = formGroups((RotSet)(rotationDiffSets.get(i)),diffs,
                                    nSym,min_group_sz);

            for(int j=0;j<tmpGroups.size();j++){
                tmpRotSet = (RotSet)tmpGroups.get(j);
                scoreRotSet(tmpRotSet,diffs,RFScoring,NCSx,NCSy,NCSz);
                if ((tmpRotSet.max_angle_err < rot_tol) && 
                    (tmpRotSet.max_axis_err < axis_tol)) {
                    if(selfAxisPresent){
                        if (tmpRotSet.max_self_axis_err < axis_tol)
                            goodGroups.add(tmpRotSet);
                    } else
                        goodGroups.add(tmpRotSet);
                }
            }
        }
        return(goodGroups);
    }

    
    // Creates preliminary sets of rotation differences with similar
    //  axes for a given vector of differences
    public Vector groupSimilarDiffs(RotDifference[][] diffs, int row,
                                        int numRots, double axis_tol) {

        Vector rotationSets = new Vector();
        RotDifference diff1 = null;
        RotDifference diff2 = null;
        RotSet rSet = null;
        
        // Start from each difference
        for(int j=0;j<numRots;j++){
            if (j != row) {
                diff1 = diffs[row][j];
                if (diff1.active) {
                    rSet = new RotSet();
                    rSet.add(diff1);
                    for(int k=(j+1);k<numRots;k++){
                        if (k != row) {
                            diff2 = diffs[row][k];
                            if (diff2.active) {
                                if ((180.0/3.1415926*Math.acos(
                                  Math.abs(diff1.axis[0]*diff2.axis[0] +
                                           diff1.axis[1]*diff2.axis[1] +
                                           diff1.axis[2]*diff2.axis[2])))
                                        < axis_tol)
                                    rSet.add(diff2);
                            }
                        }
                    }
                    rotationSets.add(rSet);
                }
            }
        }
        return(rotationSets);
    }

    
    // Returns true if rSet1 and rSet2 are equal
    public boolean group_equal(RotSet rSet1, RotSet rSet2){
        
        // Find start in rSet2
        int index1 = ((Rotation)rSet1.get(0)).index;
        int start = -1;
        int j = 0, i = 0, match = 0;
        Rotation r1 = null;
        Rotation r2 = null;
        int sz = rSet2.size();
        
        for(i=0;i<sz;i++){
            r2 = (Rotation) rSet2.get(i);
            if (r2 != null){
                if (r2.index == index1)
                    start = i;
            }
        }
        if (start == -1)
            return(false);
        
        // Check forwards
        j = start;
        match = 0;
        for(i=0;i<sz;i++){
            r1 = (Rotation)rSet1.get(i);
            r2 = (Rotation)rSet2.get(j);
            if (((r1 == null) && (r2 != null)) ||
                ((r1 != null) && (r2 == null)))
                break;
            else if ((r1 == null) && (r2 == null))
                match++;
            else if (r1.index == r2.index)
                match++;
            else
                break;
            j++;
            if (j>=sz)
                j -= sz;
        }
        if (match==sz)
            return(true);
        
        // Check backwards
        j = start;
        match = 0;
        for(i=0;i<sz;i++){
            r1 = (Rotation)rSet1.get(i);
            r2 = (Rotation)rSet2.get(j);
            if (((r1 == null) && (r2 != null)) ||
                ((r1 != null) && (r2 == null)))
                break;
            else if ((r1 == null) && (r2 == null))
                match++;
            else if (r1.index == r2.index)
                match++;
            else
                break;
            j--;
            if (j<0)
                j += sz;
        }
        if (match==sz)
            return(true);

        return(false);
    }
    
    // Returns true if we don't have group rSet in totalRS
    public boolean dontHaveGroup(RotSet rSet, Vector totalRS){
    
        for(int i=0;i<totalRS.size();i++) {
            if (group_equal(rSet,(RotSet)totalRS.get(i)))
                return(false);
        }
        return(true);
    }

    
    // Considers each set in goodRS and if it's not alrady member of 
    //  totalRS then it is added to totalRS
    public void mergeRotationSets(Vector totalRS, Vector goodRS){
    
        RotSet rSet = null;
        
        for(int i=0;i<goodRS.size();i++){
            rSet = (RotSet)goodRS.get(i);
            if(dontHaveGroup(rSet, totalRS))
                totalRS.add(rSet);
        }
    }
    
    // Computes the NCS consistent rotation sets
    public Vector computeRotationSets(Vector rotationSets, RotDifference[][] diffs,
                                      int numRots, int nSym, double rot_tol,
                                      double axis_tol, int min_group_sz,
                                      boolean RFScoring, boolean selfAxisPresent,
                                      double NCSx, double NCSy, double NCSz) {
        int foo;
        Vector tmpDiffSet = null;
        Vector goodRotSet = null;
        Vector totalRotSet = new Vector();
        
        for(int i=0;i<numRots;i++){
            tmpDiffSet = groupSimilarDiffs(diffs,i,numRots,axis_tol);
            goodRotSet = makeGroups(tmpDiffSet,diffs,i,numRots,nSym,rot_tol,
                                    axis_tol,min_group_sz,RFScoring,
                                    selfAxisPresent,NCSx,NCSy,NCSz);
            // Adds those rotation sets from goodRotSet that aren't already in
            //  totalRotSet to totalRotSet
            mergeRotationSets(totalRotSet,goodRotSet);
        }
        
        return(totalRotSet);
    }
    
    
    // 'Remove' all differences that are more than rot_tol from an
    //  i*360/nSym
    public void findCloseDifferences(RotDifference[][] diffs, int numRots, double rot_tol) {
        
        Vector tmpVec = null;
        RotDifference tmpDiff = null;
        int n=0;
        for(int i=0;i<numRots;i++){
            for(int j=0;j<numRots;j++){
                tmpDiff = diffs[i][j];
                if (tmpDiff != null) {
                    if (tmpDiff.error < rot_tol)
                        tmpDiff.active = true;
                    else {
                        tmpDiff.active = false;
                        n++;
                    }
                }
            }
        }
    }
    
    // Compute all rotation pair differences
    //  ignoring differences between rotations and their dimer
    //  related symmetry.
    public void computeAllPairsRotationDiffs(Vector rots, RotDifference[][] diffs, int nSym) {
     
        Rotation rot1 = null;
        Rotation rot2 = null;
        RotDifference rotDiff = null;
        Quaternion diffQuat = null;
        double[] axis = null;
        double angle = 0.0;
        double deltaA = 360.0 / (double)nSym;
        int sNum = 0;
        double tmpAngle = 0.0;
        double sErr = 0.0;
        for(int i=0;i<rots.size();i++){
            rot1 = (Rotation) rots.get(i);
            for(int j=0;j<rots.size();j++){
                if (i!=j) {
                    rot2 = (Rotation) rots.get(j);
                    // Proceed if these two rotations are not from the same
                    //  cross-rotation peak (ie. if they're not dimers)
                    if (!(rot1.origString.equalsIgnoreCase(rot2.origString))) {
                        diffQuat = rot2.rot.multiply(rot1.rot.conjugate());
                        axis = diffQuat.getAxis();
                        angle = diffQuat.getAngle();
                        // Find nearest sNum*360/nSym
                        sErr = 360.0;
                        tmpAngle = deltaA;
                        for(int m=1; m<nSym; m++){
                            double e = Math.abs(angle-tmpAngle);
                            if (e < sErr) {
                                sErr = e;
                                sNum = m;
                            }
                            tmpAngle += deltaA;
                        }
                        rotDiff = new RotDifference(rot1,rot2,axis,angle,sNum,sErr);
                        diffs[i][j] = rotDiff;
                    } else {
                        rotDiff = new RotDifference();
                        rotDiff.active = false;
                        diffs[i][j] = rotDiff;
                    }
                }
            }
        }
        
    }
    

    // Reads the cross-rotation peaks from the input file
    public int readCrossRotationList(String fname, Vector rots){
        
      	BufferedReader bufread = null;
	String s = new String("");  // line being parsed
	boolean done = false;
        int numRead = 0;

	// Setup and open the file
	try{
            File file = new File(fname);
            FileReader fr = new FileReader(file);
            bufread = new BufferedReader(fr);
	}
	catch (FileNotFoundException e) {
            System.out.println("ERROR: File " + fname + " Not Found");
	}

	// Read all the lines
        int rep = 0;    // Rotation representation of input file
        rep = 1; // Indicates CNS Euler Angles Being Used ZXZ
        Quaternion q = null;
        float t1 = 0.0f, t2 = 0.0f, t3 = 0.0f;
        while (!done) {
            try {
                s = bufread.readLine();
                if (s == null) { // stop if we've reached EOF
                    done = true;
                    bufread.close();
                    s = new String("");
                }
                else {
                    if (getToken(s,1).charAt(0) == '!') {
                        //if (s.indexOf("theta") >= 0) {
                        //    rep = 1;
                        //    System.out.println("CNS Euler Angles Being Used (z,x,z)");
                        //}
                        //if (s.indexOf("CCP4") >= 0) {
                        //    rep = 2;
                        //    System.out.println("CCP4 Euler Angles Being Used (z,y,z)");
                        //}
                    }
                    else {
                        q = new Quaternion();
                        t1 = (new Float(getToken(s,2)).floatValue());
                        t2 = (new Float(getToken(s,3)).floatValue());
                        t3 = (new Float(getToken(s,4)).floatValue());
                        if (rep == 1)
                            q.fromEuler(t1,t2,t3);
                        else if (rep == 2)
                            q.fromEulerCCP4(t1,t2,t3);
                        else {
                            System.out.println("ERROR: Unknown rotation representation");
                            System.exit(-1);
                        }
                        Rotation r = new Rotation(q,s.trim());
                        r.quality = (new Float(getToken(s,5)).floatValue());
                        r.index = numRead++;
                        // Add the rotation to the list of rotations
                        rots.add(r);
                    }
                }
            }
            catch ( Exception e ){
                System.out.println("ERROR: An error occured while reading input");
                System.exit(0);
            }
        }
    
        return(numRead);
    }
    
    
    // Generates the symmetry related rotations (ie. those related
    //  by a 180-degree rotation around the dimer 2-fold axis (dimerRot)
    public Vector generateDimerRelatedRots(Vector rots, Quaternion dimerQuat) {
    
        Rotation tmpRot = null;
        Rotation newRot = null;
        Vector newVec = new Vector();
        int numRots = 0;
        
        for(int i=0;i<rots.size();i++){
            tmpRot = (Rotation) rots.get(i);
            //newRot = new Rotation(dimerQuat.multiply(tmpRot.rot),tmpRot.rot,tmpRot.origString);
            newRot = new Rotation(tmpRot.rot.multiply(dimerQuat),tmpRot.rot,tmpRot.origString);
            newRot.quality = tmpRot.quality;
            tmpRot.index = numRots++;
            newRot.index = numRots++;
            newVec.add(tmpRot);
            newVec.add(newRot);
        }
        
        rots = newVec;
        return(rots);
    }
    
    
    // This function returns the xth token in string s
    private String getToken(String s, int x) {

      int curNum = 1;	
      StringTokenizer st = new StringTokenizer(s," ,;\t\n\r\f");

      while (curNum < x) {
        curNum++;
        if (st.hasMoreTokens())
            st.nextToken();
        else {
            // System.out.println("ERROR: Unable to access parameter " + x);
            return(new String(""));
        }
      }

      if (st.hasMoreTokens())
        return(st.nextToken());
      return(new String(""));

    } // end getToken

}
