/*
 * GetDimerAxis.java
 *
 * This class computes a 2-fold dimer axis
 *  from a pdb file containing the dimer.
 * The axis is computed from all pairwise
 *  CA atoms
 *
 *  "A Subgroup Algorithm to Identify Cross-Rotation Peaks
 *   Consistent with Non-Crystallographic Symmetry." R. H. Lilien,
 *   C. Bailey-Kellogg, A. C. Anderson, and B. R. Donald, Acta
 *   Crystallographica D, 60:1057-67 (2004).
 *
 * Ryan Lilien
 * Chris Bailey-Kellogg

////////////////////////////////////////////////////////////////////////////////////////////
//  GetDimerAxis.java
//
//  Version:           0.1
//
//
// authors:
//    initials    name            organization                email
//   ---------   --------------  ------------------------    ------------------------------
//     RHL        Ryan Lilien     Dartmouth College           lilien@cs.dartmouth.edu
//
////////////////////////////////////////////////////////////////////////////////////////////

*/

package CRANS;

import java.util.*;
import java.io.*;
import java.lang.*;


public class GetDimerAxis {
    
    public static final int MAXPAIRS = 10000;
    
    HashMap dimA = null;  // The heap for CA of dimer chain A
    HashMap dimB = null;  // The heap for CA of dimer chain B
    HashMap axes = null;  // The heap of all computed dimer axes
    int maxResNum = 0;    // The maximum residue number seen
   
    // The constructor reads the pdb and parses the CA atoms into
    //  each chain hashmap.
    public GetDimerAxis(String fname, String chainA, String chainB) {
        
        // Make sure we can read the file
        BufferedReader bufread = null;
	String s = new String("");  // line being parsed
	boolean done = false;
        Integer resnum = null;
        String coords = null;
        String chainID = null;

	// Setup and open the file
	try{
            File file = new File(fname);
            FileReader fr = new FileReader(file);
            bufread = new BufferedReader(fr);
	}
	catch (Exception e) {
            System.out.println("ERROR: Error reading dimer pdb file: " + fname);
	}

        dimA = new HashMap(100);
        dimB = new HashMap(100);
        maxResNum = 0;

        // Read all the lines
        while (!done) {
            try {
                s = bufread.readLine();
                if (s == null) { // stop if we've reached EOF
                    done = true;
                    bufread.close();
                    s = new String("");
                }
                else {
                    if (getToken(s,1).equalsIgnoreCase("ATOM")) {
                        coords = new String("");
                        if (s.substring(13,15).equalsIgnoreCase("CA")) {
                            resnum = new Integer(s.substring(22,26).trim());
                            coords = s.substring(30,54);
                            chainID = s.substring(21,22);
                            if (!chainID.equalsIgnoreCase(chainA) && !chainID.equalsIgnoreCase(chainB))
                                chainID = s.substring(72,76).trim();
                            if (chainID.equalsIgnoreCase(chainA))
                                dimA.put(resnum,coords);
                            else if (chainID.equalsIgnoreCase(chainB))
                                dimB.put(resnum,coords);
                            if (resnum.intValue() > maxResNum)
                                maxResNum = resnum.intValue();
                        }
                    }
                }
            }
            catch ( Exception e ){
                System.out.println("ERROR: An error occured while reading dimer file");
                System.exit(0);
            }
        }

    }

    
    // Computes the dimer axis by computing the cross product
    //  between n^2 pairs of atoms (n = number of residues).
    // Actually, max(n^2,MAXPAIRS) pairs are examined
    public double[] computeAxis() {
        
        int numpairs = 0;
        Integer theKey = null;
        String vecA1 = null;
        String vecB1 = null;
        String vecA2 = null;
        String vecB2 = null;
        double[] atA1 = new double[3];
        double[] atB1 = new double[3];
        double[] atA2 = new double[3];
        double[] atB2 = new double[3];
        double[] vec1 = new double[3];
        double[] vec2 = new double[3];
        double[] vec3 = new double[3];
        double[] axis = new double[3];
    
        if ((dimA == null) || (dimB == null))
            return(null);
        
        for(int q=0; q<=maxResNum; q++){
            theKey = new Integer(q);
            vecA1 = (String)dimA.get((Object)theKey);
            vecB1 = (String)dimB.get((Object)theKey);
            if ((vecA1 != null) && (vecB1 != null)) {
                for(int w=(q+1); w<=maxResNum; w++){
                    theKey = new Integer(w);
                    vecA2 = (String)dimA.get((Object)theKey);
                    vecB2 = (String)dimB.get((Object)theKey);
                    if ((vecA2 != null) && (vecB2 != null)) {
                        atA1 = parseCoords(vecA1);
                        atB1 = parseCoords(vecB1);
                        atA2 = parseCoords(vecA2);
                        atB2 = parseCoords(vecB2);
                        vec1[0] = atA1[0] - atB1[0];
                        vec1[1] = atA1[1] - atB1[1];
                        vec1[2] = atA1[2] - atB1[2];
                        vec2[0] = atA2[0] - atB2[0];
                        vec2[1] = atA2[1] - atB2[1];
                        vec2[2] = atA2[2] - atB2[2];
                        vec1 = norm(vec1);
                        vec2 = norm(vec2);
                        vec3 = cross(vec1,vec2);
                        vec3 = norm(vec3);
                        if (vec3[0] < 0.0){
                            vec3[0] = -vec3[0];
                            vec3[1] = -vec3[1];
                            vec3[2] = -vec3[2];
                        }
                        axis[0] += vec3[0];
                        axis[1] += vec3[1];
                        axis[2] += vec3[2];
                        numpairs++;
                        if (numpairs >= MAXPAIRS) {
                            w = maxResNum+1;
                            q = w;
                        }
                    }
                }
            }
        }
        
        axis[0] /= numpairs;
        axis[1] /= numpairs;
        axis[2] /= numpairs;
        return(axis);
    }
    
    
    // Parses the 3 values in str to a double array
    private double[] parseCoords(String str) {

        double[] da = new double[3];
        da[0] = (new Double(getToken(str,1))).doubleValue();
        da[1] = (new Double(getToken(str,2))).doubleValue();
        da[2] = (new Double(getToken(str,3))).doubleValue();
        return(da);
    }
    
    // Normalizes vector v
    private double[] norm(double[] v) {
        
        double len = 0.0;
        double[] nv = new double[3];
        len = Math.sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
        nv[0] = v[0]/len;
        nv[1] = v[1]/len;
        nv[2] = v[2]/len;
        return(nv);
    }
    
    // Computes the cross product between vector 1 and 2
    private double[] cross(double[] v1, double[] v2) {
        
        double[] nv = new double[3];
        nv[0] = v1[1]*v2[2] - v1[2]*v2[1];
        nv[1] = v1[2]*v2[0] - v1[0]*v2[2];
        nv[2] = v1[0]*v2[1] - v1[1]*v2[0];
        return(nv);
    }
    
    // This function returns the xth token in string s
    private String getToken(String s, int x) {

    int curNum = 1;	
    StringTokenizer st = new StringTokenizer(s," ,;\t\n\r\f");

    while (curNum < x) {
        curNum++;
        if (st.hasMoreTokens())
            st.nextToken();
        else {
            // System.out.println("ERROR: Unable to access parameter " + x);
            return(new String(""));
        }
    }

    if (st.hasMoreTokens())
        return(st.nextToken());
    return(new String(""));

    } // end getToken

}
