/*
 * Quaternion.java
 *
 * This class implements the quaternion class utilized by the
 *  CRANS algorithm.
 *
 *  "A Subgroup Algorithm to Identify Cross-Rotation Peaks
 *   Consistent with Non-Crystallographic Symmetry." R. H. Lilien,
 *   C. Bailey-Kellogg, A. C. Anderson, and B. R. Donald, Acta
 *   Crystallographica D, 60:1057-67 (2004).
 *
 * Ryan Lilien
 * Chris Bailey-Kellogg

////////////////////////////////////////////////////////////////////////////////////////////
//  Quaternion.java
//
//  Version:           0.1
//
//
// authors:
//    initials    name            organization                email
//   ---------   --------------  ------------------------    ------------------------------
//     RHL        Ryan Lilien     Dartmouth College           lilien@cs.dartmouth.edu
//
////////////////////////////////////////////////////////////////////////////////////////////

*/


package CRANS;

import java.lang.*;

// All angles are in degrees

public class Quaternion {

    // Conversion constants
    public static final double D2R = Math.PI / 180.0;
    public static final double R2D = 180.0 / Math.PI;
    
    double[] vec = new double[3];   // quaternion vector
    double s;                       // quaternion scalar
    
    // Creates an empty zero quaternion
    public Quaternion() {
        vec[0] = 0.0;
        vec[1] = 0.0;
        vec[2] = 0.0;        
        s = 0.0;
    }
   
    // Creates a quaternion from the input vector and scalar
    public Quaternion(double[] tmpV, double tmpS) {
        vec[0] = tmpV[0];
        vec[1] = tmpV[1];
        vec[2] = tmpV[2];
        s = tmpS;
    }

    public void normalize() {
        double len = Math.sqrt(vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2] + s*s);
        vec[0] /= len;
        vec[1] /= len;
        vec[2] /= len;
        s /= len;
    }
    
    // Returns the conjugate
    public Quaternion conjugate() {
        Quaternion q = new Quaternion();
        q.s = s;
        q.vec[0] = -vec[0];
        q.vec[1] = -vec[1];
        q.vec[2] = -vec[2];
        return(q);
    }
    
    // Returns the result of multiplying q2*this_quaternion
    public Quaternion multiply(Quaternion q2) {
        
        Quaternion q3 = new Quaternion();
        q3.s = (q2.s * s) - (q2.vec[0]*vec[0] + q2.vec[1]*vec[1] + q2.vec[2]*vec[2]);
        q3.vec[0] = q2.s*vec[0] + s*q2.vec[0] + (q2.vec[1]*vec[2] - q2.vec[2]*vec[1]);
        q3.vec[1] = q2.s*vec[1] + s*q2.vec[1] + (q2.vec[2]*vec[0] - q2.vec[0]*vec[2]);
        q3.vec[2] = q2.s*vec[2] + s*q2.vec[2] + (q2.vec[0]*vec[1] - q2.vec[1]*vec[0]);
        q3.normalize();
        return(q3);
    }

    
    // Returns the axis of the axis/angle reprenstation of this quaterion
    public double[] getAxis() {
        
        double sa = 0.0;
        double[] theAxis = null;
        
        normalize();
        sa = Math.sqrt(1.0 - s*s);
        theAxis = new double[3];
        if (Math.abs(sa) < 0.00001){
            theAxis[0] = 0.0;
            theAxis[1] = 0.0;
            theAxis[2] = 0.0;
        }
        else {
            theAxis[0] = vec[0] / sa;
            theAxis[1] = vec[1] / sa;
            theAxis[2] = vec[2] / sa;
        }
        return(theAxis);
    }
    
    
    // Returns the angle of the axis/angle reprenstation of this quaterion
    public double getAngle() {
        
        double theAngle = 0.0;
        
        theAngle = Math.acos(s) * 2.0;
        theAngle *= R2D;
        return(theAngle);
    }
    
    
    // Set the value of this quaternion from an axis and angle
    public void fromAxisAngle(double axis[], double angle) {
        double a = angle*D2R / 2.0;
        double sa = Math.sin(a);
        vec[0] = axis[0]*sa;
        vec[1] = axis[1]*sa;
        vec[2] = axis[2]*sa;
        s = Math.cos(a);
        normalize();
    }

  
    // Set the value of this quaternion from Euler (z,x,z)
    public void fromEuler(double t1, double t2, double t3) {

        double[] zv = {0.0, 0.0, 1.0};
        double[] xv = {1.0, 0.0, 0.0};
        Quaternion zq1 = new Quaternion();
        Quaternion xq = new Quaternion();
        Quaternion zq2 = new Quaternion();
        zq1.fromAxisAngle(zv, t1);
        xq.fromAxisAngle(xv, t2);
        zq2.fromAxisAngle(zv, t3);
        
        // Compose them (reverse order)
        Quaternion q1 = xq.multiply(zq1);
        Quaternion q2 = zq2.multiply(q1);
        
        vec = q2.vec;
        s = q2.s;
    }
    

    // Set the value of this quaternion from Euler (z,y,z)
    public void fromEulerCCP4(double t1, double t2, double t3) {

        double[] zv = {0.0, 0.0, 1.0};
        double[] yv = {0.0, 1.0, 0.0};
        Quaternion zq1 = new Quaternion();
        Quaternion yq = new Quaternion();
        Quaternion zq2 = new Quaternion();
        zq1.fromAxisAngle(zv, t1);
        yq.fromAxisAngle(yv, t2);
        zq2.fromAxisAngle(zv, t3);
        
        // Compose them (reverse order)
        Quaternion q1 = yq.multiply(zq1);
        Quaternion q2 = zq2.multiply(q1);
        
        vec = q2.vec;
        s = q2.s;
    }

    // Converts this quaternion to a rotation matrix
    public double[][] toMatrix() {
        
        double xx = vec[0]*vec[0];
        double xy = vec[0]*vec[1];
        double xz = vec[0]*vec[2];
        double xw = vec[0]*s;
        
        double yy = vec[1]*vec[1];
        double yz = vec[1]*vec[2];
        double yw = vec[1]*s;
        
        double zz = vec[2]*vec[2];
        double zw = vec[2]*s;
        
        double[][] m = new double[3][3];
        m[0][0] = 1.0 - 2.0 * (yy+zz);
        m[0][1] = 2.0 * (xy-zw);
        m[0][2] = 2.0 * (xz+yw);
        
        m[1][0] = 2.0 * (xy+zw);
        m[1][1] = 1.0 - 2.0 * (xx+zz);
        m[1][2] = 2.0 * (yz-xw);
        
        m[2][0] = 2.0 * (xz-yw);
        m[2][1] = 2.0 * (yz+xw);
        m[2][2] = 1.0 - 2.0 * (xx+yy);
        return(m);
    }
    
    // Converts the quaternion into z,x,z euler angles and prints
    public double[] toEuler() {
    
        // Find rotation matrix and extract z,x,z
        // If Z1 = [a -b 0; b a 0; 0 0 1]
        //     X = [1 0 0; 0 c -d; 0 d c]
        //    Z2 = [e -f 0; f e 0; 0 0 1]
        // Then M = [ae-bcf -af-bce bd; be+acf -bf+ace -ad; df de c]
        
        double t1,t2,t3;
        double[][] m = toMatrix();
        double a,b,c,d,e,f;
        c = m[2][2];
        t2 = Math.acos(c);
        d = Math.sin(t2);
        if (Math.abs(d) < 0.00001) {
            t1 = Math.atan2(m[1][0],m[1][1]);
            t3 = 0.0;
        } else {
            a = -m[1][2]/d;
            b = m[0][2]/d;
            t1 = Math.atan2(b,a);
            e = m[2][1]/d;
            f = m[2][0]/d;
            t3 = Math.atan2(f,e);
        }
        
        if (t1<0)
            t1 += 2.0*Math.PI;
        if (t2<0)
            t2 += 2.0*Math.PI;
        if (t3<0)
            t3 += 2.0*Math.PI;
       
        // Convert back to degrees
        t1 *= R2D;
        t2 *= R2D;
        t3 *= R2D;
        
        double[] t = new double[3];
        t[0] = t1;
        t[1] = t2;
        t[2] = t3;
        return(t);
    }

    // Converts the quaternion into z,x,z euler angles and prints
    public String toEulerString() {

        double[] t = toEuler();
        String s = new String(t[0]+" "+t[1]+" "+t[2]);
        return(s);
    }
    
}
