/*
 * RotSet.java
 *
 * Rotation groups in the CRANS algorithm are represented
 *  by this class (RotSet).
 *
 *  "A Subgroup Algorithm to Identify Cross-Rotation Peaks
 *   Consistent with Non-Crystallographic Symmetry." R. H. Lilien,
 *   C. Bailey-Kellogg, A. C. Anderson, and B. R. Donald, Acta
 *   Crystallographica D, 60:1057-67 (2004).
 *
 * Ryan Lilien
 * Chris Bailey-Kellogg

////////////////////////////////////////////////////////////////////////////////////////////
//  RotSet.java
//
//  Version:           0.1
//
//
// authors:
//    initials    name            organization                email
//   ---------   --------------  ------------------------    ------------------------------
//     RHL        Ryan Lilien     Dartmouth College           lilien@cs.dartmouth.edu
//
////////////////////////////////////////////////////////////////////////////////////////////

*/

package CRANS;

import java.util.*;
import java.lang.Comparable;

// This is a set of rotations

public class RotSet implements Comparable {
    
    Vector rots = new Vector();
    double score = 0.0;
    double mean_axis_err = 0.0;
    double max_axis_err = 0.0;
    double mean_angle_err = 0.0;
    double max_angle_err = 0.0;
    double max_self_axis_err = 0.0;
    
    public RotSet() {
    }
    
    public RotSet(RotSet oldRS){
     
        // Copy rots elements
        rots = (Vector) oldRS.rots.clone();
        score = oldRS.score;
        mean_axis_err = oldRS.mean_axis_err;
        max_axis_err = oldRS.max_axis_err;
        mean_angle_err = oldRS.mean_angle_err;
        max_angle_err = oldRS.max_angle_err;
    }
    
    // Adds the rotation ob to this rotation set
    public void add(Object ob){
        rots.add(ob);
    }

    // Returns the rotation at index idx
    public Object get(int idx){
        return(rots.get(idx));
    }
    
    // Sets the size of the set
    public void setSize(int sz){
        rots.setSize(sz);
    }
    
    // Changes element n of the rotation set
    public void set(int n, Object ob){
        rots.setElementAt(ob,n);
    }
    
    // Returns the size of the rotation set
    public int size(){
        return(rots.size());
    }
    
    // Returns the number of rotations in the set which are not simulated
    public int numTrueRotations(){
        int n=0;
        Rotation tmpRot = null;
        for(int i=0;i<rots.size();i++){
            tmpRot = (Rotation)rots.get(i);
            if (tmpRot != null)
                if (!(tmpRot.simulated))
                    n++;
        }
        return(n);
    }
    
    // Used in comparing rotation sets for sorting
    public int compareTo(Object tmpObj) {
    
        RotSet tmpRS = (RotSet) tmpObj;
        int nTR = numTrueRotations();
        int onTR = tmpRS.numTrueRotations();
        // First check which set has more natural rotations
        if(nTR > onTR)
            return(-1);
        else if (nTR < onTR)
            return(1);
        
        // If they have the same number of natural rotations, examine
        //  the rotation set score
        if ((score - tmpRS.score) < 0.0)
            return(1);
        else if ((score - tmpRS.score) > 0.0)
            return(-1);
        else
            return(0);
    }
    
}
